/* @(#) get friends from LJ */

#include	<stdio.h>
#include	<string.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<netdb.h>

#define		HTTPPORT	80	/* HTTP port */

#define		OK		0	/* okay return value */
#define		ERROR		(-1)	/* error return value */

#define		ARGCHAR		'-'	/* argument delimiter */
#define		PATHSEP		'/'	/* path separator */

#define		FALSE		0	/* boolean `false' */
#define		TRUE		1	/* boolean `true' */

static int	gethost(struct in_addr * iap, const char * host);
static int	opennet(struct in_addr * iap, int port);
static void	douser(char * user);
static void	doljcmd(const char * fmt, ...);
static void	doline(int header, char * ptr, char * end);
static void	sendfd(int fd, void * ptr, int len);
static void	closenet(int sfd);
static void	setprog(char * name);
static void	dprint(const char * fmt, ...);
static void	error(const char * fmt, ...);
static void	whyerror(const char * fmt, ...);
static void *	mkbuf(int size);

static char *	Usage      = "usage: %s [-L ljhost] [-P proxy] <user>";
static char *	Getfriends = "misc/fdata.bml?user=%s";
static char *	Ljhost     = "www.livejournal.com";
static char *	Eol        = "\r\n";
static char *	Write      = "w";

static int		debugflag;	/* debug flag */
static char *		proxyhost;	/* use HTTP proxy */
static char *		prog;		/* program name */
static FILE *		ofp;		/* output file pointer */
static struct in_addr	ljhost;		/* LiveJournal host */

int
main(
	int		argc,		/* argument count */
	char **		argv,		/* argument vector */
	char **		envp)		/* environment pointer */
{
	char *		va;		/* value array */
	char *		nvp;		/* needed value pointer */
	char *		cvp;		/* current value pointer */
	char *		evp;		/* end of value pointers */
	char *		user;		/* LJ username */

	setprog(*argv);

	va            = mkbuf(argc);
	nvp           = va;
	cvp           = va;
	evp           = va + argc;
	ofp           = stdout;
	ljhost.s_addr = 0;
	user          = NULL;
	proxyhost     = NULL;
	debugflag     = FALSE;

	while (--argc) {
		if (**++argv == ARGCHAR) {
			while (*++*argv) {
				switch (**argv) {
					case 'm':	/* max friends */
					case 'o':	/* output file */
					case 'L':	/* specify host */
					case 'P':	/* use proxy */
						if (nvp == evp) {
							error(Usage, prog);
							}

						*nvp++ = **argv;
						break;

					case 'D':
						debugflag = TRUE;
						break;

					default:
						error(Usage, prog);
					}
				}
			}
		else {
			if (nvp > cvp) {
				switch (*cvp++) {
					case 'o':	/* output file */
						ofp = fopen(*argv, Write);

						if (ofp == NULL) {
							whyerror(*argv);
							}
						break;

					case 'L':	/* specify host */
						Ljhost = *argv;
						break;

					case 'P':	/* proxy host */
						proxyhost = *argv;
						break;
					}
				}
			else {
				user = *argv;
				}
			}
		}

	free(va);

	if (user == NULL) {
		error(Usage, prog);
		}

	fprintf(ofp, "<?xml version=\"1.0\"?>\n");
	fprintf(ofp, "<friends>\n");

	douser(user);

	fprintf(ofp, "</friends>\n");

	return	0;
}

/* process a user */

static void
douser(
	char *	user)	/* user to process */
{
	fprintf(ofp, "  <user name=\"%s\">\n", user);

	doljcmd(Getfriends, user);

	fprintf(ofp, "  </user>\n");
}

/* go get the information */

static void
doljcmd(
	const char *		fmt,		/* query format */
				...)		/* data pointer */
{
	int			nxfer;		/* # of bytes transferred */
	int			gotnl;		/* got newline */
	int			header;		/* header line flag */
	int			sfd;		/* socket file descriptor */
	int			ncb;		/* number of command bytes */
	int			nhb;		/* number of header bytes */
	char			hdr[BUFSIZ];	/* header buffer */
	char			cmd[BUFSIZ];	/* command buffer */
	char			netbuf[BUFSIZ];	/* return data buffer */
	char *			host;		/* host name */
	char *			lbp;		/* last buffer pointer */
	char *			nbp;		/* network buffer pointer */
	char *			nbe;		/* network buffer end */
	char *			bsp;		/* buffer start pointer */
	va_list			ap;		/* argument pointer */

	/* build command string */

	va_start(ap, fmt);

	ncb = vsprintf(cmd, fmt, ap);

	va_end(ap);

	/* build header */

	nhb = sprintf(hdr, "GET /%s HTTP/1.0%s", cmd, Eol);
	nhb += sprintf(hdr + nhb,
		"User-Agent: http://www.vitriol.com; webmaster@vitriol.com%s",
		Eol);

	nhb += sprintf(hdr + nhb, "%s", Eol);

	if (ljhost.s_addr == 0) {
		host = Ljhost;

		if (gethost(&ljhost, host)) {
			whyerror(host);
			}
		}

	dprint("opening network connection to `%s'\n", host);

	sfd = opennet(&ljhost, HTTPPORT);

	if (sfd < 0) {
		whyerror(host);
		}

	dprint("sending request\n");

	sendfd(sfd, hdr, nhb);

	dprint("getting reply\n");

	header = TRUE;
	gotnl  = FALSE;
	bsp    = netbuf;

	while ((nxfer = read(sfd, bsp, BUFSIZ - (bsp - netbuf))) > 0) {
		dprint("got %d bytes\n", nxfer);
		lbp = netbuf;
		nbe = bsp + nxfer;

		for (nbp = lbp; nbp < nbe; nbp++) {
			switch (*nbp) {
				case '\r':
					dprint("got cr");
					break;

				case '\n':
					dprint("got nl");
					if (gotnl) {
						header = FALSE;
						}
					else {
						gotnl = TRUE;

						if (lbp != nbp) {
							doline(header,
								lbp, nbp);
							}
						}
					break;

				default:
					if (gotnl) {
						lbp   = nbp;
						gotnl = FALSE;
						}
					break;
				}
			}

		if (gotnl) {
			bsp   = netbuf;
			gotnl = FALSE;
			}
		else {
			dprint("keeping %d bytes", nbe - lbp);
			(void) memmove(netbuf, lbp, nbe - lbp);
			bsp = netbuf + (nbe - lbp);
			}
		}

	dprint("closing connection\n");

	closenet(sfd);
}

/* deal with a line from the server */

static void
doline(
	int	header,	/* header flag */
	char *	ptr,	/* data start */
	char *	end)	/* data end */
{
	int	type;		/* type of record */
	char	buf[32];	/* name buffer */
	char *	bp;		/* buffer pointer */

	dprint("doline(): got %d characters", end - ptr);

	if (debugflag) {
		char *	pc;

		if (header) {
			fprintf(stderr, "(header) ");
			}

		for (pc = ptr; pc < end; pc++) {
			putc(*pc, stderr);
			}
		}

	if (header) {
		return;
		}

	type = *ptr;

	if (type == '#') {
		return;
		}

	bp   = buf;
	ptr += 2;

	while (ptr < end) {
		*bp++ = *ptr++;
		}

	*bp = '\0';

	switch (type) {
		case '>':
			fprintf(ofp, "    <friend>%s</friend>\n", buf);
			break;

		case '<':
			fprintf(ofp, "    <friendof>%s</friendof>\n", buf);
			break;
		}
}

/* send data to a file descriptor */

static void
sendfd(
	int	fd,	/* file descriptor */
	void *	ptr,	/* data pointer */
	int	len)	/* data length */
{
	int	nxfer;	/* bytes transferred */

	dprint("sending %d bytes", len);

	nxfer = write(fd, ptr, len);

	if (nxfer != len) {
		whyerror("sent %d of %d bytes", nxfer, len);
		}
}

/* get host address */

static int
gethost(
	struct in_addr *	iap,		/* internet address pointer */
	const char *		host)		/* host name or dotted quad */
{
	struct hostent *	hep;		/* host entry pointer */

	if (!*host) {
		return	0;
		}

	hep = gethostbyname(host);

	if (hep != NULL) {
		(void) memcpy(&iap->s_addr, hep->h_addr, sizeof(*iap));
		return	OK;
		}

	iap->s_addr = inet_addr(host);

	if (iap->s_addr != (unsigned) (ERROR)) {
		return	OK;
		}

	return	ERROR;
}

/* open network connection */

static int
opennet(
	struct in_addr *	iap,		/* host */
	int			port)		/* port */
{
	int			fd;		/* file descriptor */
	struct sockaddr_in	sa;		/* INET socket address */

	memset((void *) &sa, 0, sizeof(sa));

	fd = socket(PF_INET, SOCK_STREAM, 0);

	if (fd < 0) {
		return	ERROR;
		}

	sa.sin_family      = AF_INET;
	sa.sin_addr.s_addr = iap->s_addr;
	sa.sin_port        = port;

	dprint("connecting to %s:%d", inet_ntoa(*iap), sa.sin_port);

	if (connect(fd, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
		return	ERROR;
		}

	return	fd;
}

/* close network connection */

static void
closenet(
	int	sfd)		/* socket file descriptor */
{
	(void) close(sfd);
}

/* set currently running program name */

static void
setprog(
	char *	name)	/* program name to use */
{
	prog = strrchr(name, PATHSEP);

	if (prog) {
		prog++;
		}
	else {
		prog = name;
		}
}

/* print out message if debug enabled */

static void
dprint(
	const char *	fmt,	/* message format */
			...)	/* format arguments */
{
	va_list		ap;	/* argument pointer */

	if (!debugflag) {
		return;
		}

	fflush(stdout);

	fprintf(stderr, "%s: ", prog);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	fprintf(stderr, "\n");
}

/* print out message and exit */

static void
error(
	const char *	fmt,	/* message format */
			...)	/* format arguments */
{
	va_list		ap;	/* argument pointer */

	fflush(stdout);

	fprintf(stderr, "%s: ", prog);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	fprintf(stderr, "\n");

	exit(EINVAL);
}

/* print out error message and reason and exit */

static void
whyerror(
	const char *	fmt,	/* message format */
			...)	/* format arguments */
{
	char *		errstr;	/* error string pointer */
	va_list		ap;	/* argument pointer */

	fflush(stdout);

	fprintf(stderr, "%s: ", prog);

	va_start(ap, fmt);

	vfprintf(stderr, fmt, ap);

	va_end(ap);

	errstr = strerror(errno);

	if (errstr) {
		fprintf(stderr, ": %s\n", errstr);
		}
	else {
		fprintf(stderr, ": errno=%d\n", errno);
		}

	if (errno) {
		exit(errno);
		}
	else {
		exit(EINVAL);
		}
}

/* allocate a buffer and punt on failure */

static void *
mkbuf(
	int	size)		/* size of buffer to allocate */
{
	void *	ptr;		/* address of new buffer */

	ptr = malloc(size);

	if (ptr == NULL) {
		whyerror("Malloc(%d)", size);
		}

	return	ptr;
}
